@@ -27,10 +27,11 @@
 #include <string.h>
 #include <fcntl.h>
 #include <unistd.h>
-#include <error.h>
+#include <err.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 
+#include "asm/insn.h"
 #include "kpatch-elf.h"
 
 /*******************
@@ -53,18 +54,18 @@ char *status_str(enum status status)
 	return NULL;
 }
 
-int is_rela_section(struct section *sec)
+bool is_rela_section(struct section *sec)
 {
 	return (sec->sh.sh_type == SHT_RELA);
 }
 
-int is_text_section(struct section *sec)
+bool is_text_section(struct section *sec)
 {
 	return (sec->sh.sh_type == SHT_PROGBITS &&
 		(sec->sh.sh_flags & SHF_EXECINSTR));
 }
 
-int is_debug_section(struct section *sec)
+bool is_debug_section(struct section *sec)
 {
 	char *name;
 	if (is_rela_section(sec))
@@ -132,6 +133,21 @@ struct rela *find_rela_by_offset(struct section *relasec, unsigned int offset)
 	return NULL;
 }
 
+unsigned int absolute_rela_type(struct kpatch_elf *kelf)
+{
+	switch(kelf->arch) {
+	case PPC64:
+		return R_PPC64_ADDR64;
+	case X86_64:
+		return R_X86_64_64;
+	case S390:
+		return R_390_64;
+	default:
+		ERROR("unsupported arch");
+	}
+	return 0;
+}
+
 /* returns the offset of the string in the string table */
 int offset_of_string(struct list_head *list, char *name)
 {
@@ -142,7 +158,7 @@ int offset_of_string(struct list_head *list, char *name)
 	list_for_each_entry(string, list, list) {
 		if (!strcmp(string->name, name))
 			return index;
-		index += strlen(string->name) + 1;
+		index += (int)strlen(string->name) + 1;
 	}
 
 	/* allocate a new string */
@@ -151,65 +167,183 @@ int offset_of_string(struct list_head *list, char *name)
 	return index;
 }
 
-void kpatch_create_rela_list(struct kpatch_elf *kelf, struct section *sec)
+static void rela_insn(const struct section *sec, const struct rela *rela,
+		      struct insn *insn)
 {
-	int rela_nr, index = 0, skip = 0;
+	unsigned long insn_addr, start, end, rela_addr;
+
+	start = (unsigned long)sec->data->d_buf;
+	end = start + sec->sh.sh_size;
+
+	if (end <= start)
+		ERROR("bad section size");
+
+	rela_addr = start + rela->offset;
+	for (insn_addr = start; insn_addr < end; insn_addr += insn->length) {
+		insn_init(insn, (void *)insn_addr, 1);
+		insn_get_length(insn);
+		if (!insn->length)
+			ERROR("can't decode instruction in section %s at offset 0x%lx",
+			      sec->name, insn_addr);
+		if (rela_addr >= insn_addr &&
+		    rela_addr < insn_addr + insn->length)
+			return;
+	}
+
+	ERROR("can't find instruction for rela at %s+0x%x",
+	      sec->name, rela->offset);
+}
+
+/*
+ * Return the addend, adjusted for any PC-relative relocation trickery, to
+ * point to the relevant symbol offset.
+ */
+long rela_target_offset(struct kpatch_elf *kelf, struct section *relasec,
+			struct rela *rela)
+{
+	long add_off;
+	struct section *sec = relasec->base;
+
+	switch(kelf->arch) {
+	case PPC64:
+		add_off = 0;
+		break;
+	case X86_64:
+		if (!is_text_section(sec) ||
+		    rela->type == R_X86_64_64 ||
+		    rela->type == R_X86_64_32S)
+			add_off = 0;
+		else if (rela->type == R_X86_64_PC32 ||
+			 rela->type == R_X86_64_PLT32 ||
+			 rela->type == R_X86_64_NONE) {
+			struct insn insn;
+			rela_insn(sec, rela, &insn);
+			add_off = (long)insn.next_byte -
+				  (long)sec->data->d_buf -
+				  rela->offset;
+		} else
+			ERROR("unhandled rela type %d", rela->type);
+		break;
+	case S390:
+		/*
+		 * For branch and relative load instructions,
+		 * add_off is -2.
+		 */
+		if (rela->type == R_390_GOTENT ||
+			rela->type == R_390_PLT32DBL ||
+			rela->type == R_390_PC32DBL)
+			add_off = -2;
+		else if (rela->type == R_390_32 ||
+			rela->type == R_390_64 ||
+			rela->type == R_390_PC32 ||
+			rela->type == R_390_PC64)
+			add_off = 0;
+		else
+			ERROR("unhandled rela type %d", rela->type);
+		break;
+	default:
+		ERROR("unsupported arch\n");
+	}
+
+	return rela->addend + add_off;
+}
+
+unsigned int insn_length(struct kpatch_elf *kelf, void *addr)
+{
+	struct insn decoded_insn;
+	char *insn = addr;
+
+	switch(kelf->arch) {
+
+	case X86_64:
+		insn_init(&decoded_insn, addr, 1);
+		insn_get_length(&decoded_insn);
+		return decoded_insn.length;
+
+	case PPC64:
+		return 4;
+
+	case S390:
+		switch(insn[0] >> 6) {
+		case 0:
+			return 2;
+		case 1:
+		case 2:
+			return 4;
+		case 3:
+			return 6;
+		}
+
+	default:
+		ERROR("unsupported arch");
+	}
+
+	return 0;
+}
+
+static void kpatch_create_rela_list(struct kpatch_elf *kelf,
+				    struct section *relasec)
+{
+	int index = 0, skip = 0;
 	struct rela *rela;
 	unsigned int symndx;
+	unsigned long rela_nr;
 
 	/* find matching base (text/data) section */
-	sec->base = find_section_by_index(&kelf->sections, sec->sh.sh_info);
-	if (!sec->base)
-		ERROR("can't find base section for rela section %s", sec->name);
+	relasec->base = find_section_by_index(&kelf->sections, relasec->sh.sh_info);
+	if (!relasec->base)
+		ERROR("can't find base section for rela section %s", relasec->name);
 
 	/* create reverse link from base section to this rela section */
-	sec->base->rela = sec;
+	relasec->base->rela = relasec;
 
-	rela_nr = sec->sh.sh_size / sec->sh.sh_entsize;
+	rela_nr = relasec->sh.sh_size / relasec->sh.sh_entsize;
 
-	log_debug("\n=== rela list for %s (%d entries) ===\n",
-		sec->base->name, rela_nr);
+	log_debug("\n=== rela list for %s (%ld entries) ===\n",
+		relasec->base->name, rela_nr);
 
-	if (is_debug_section(sec)) {
+	if (is_debug_section(relasec)) {
 		log_debug("skipping rela listing for .debug_* section\n");
 		skip = 1;
 	}
 
 	/* read and store the rela entries */
 	while (rela_nr--) {
-		ALLOC_LINK(rela, &sec->relas);
+		ALLOC_LINK(rela, &relasec->relas);
 
-		if (!gelf_getrela(sec->data, index, &rela->rela))
+		if (!gelf_getrela(relasec->data, index, &rela->rela))
 			ERROR("gelf_getrela");
 		index++;
 
 		rela->type = GELF_R_TYPE(rela->rela.r_info);
 		rela->addend = rela->rela.r_addend;
-		rela->offset = rela->rela.r_offset;
-		symndx = GELF_R_SYM(rela->rela.r_info);
+		rela->offset = (unsigned int)rela->rela.r_offset;
+		symndx = (unsigned int)GELF_R_SYM(rela->rela.r_info);
 		rela->sym = find_symbol_by_index(&kelf->symbols, symndx);
 		if (!rela->sym)
 			ERROR("could not find rela entry symbol\n");
 		if (rela->sym->sec &&
 		    (rela->sym->sec->sh.sh_flags & SHF_STRINGS)) {
-			rela->string = rela->sym->sec->data->d_buf + rela->addend;
+			rela->string = rela->sym->sec->data->d_buf +
+				       rela->sym->sym.st_value +
+				       rela_target_offset(kelf, relasec, rela);
 			if (!rela->string)
-				ERROR("could not lookup rela string for %s+%d",
+				ERROR("could not lookup rela string for %s+%ld",
 				      rela->sym->name, rela->addend);
 		}
 
 		if (skip)
 			continue;
-		log_debug("offset %d, type %d, %s %s %d", rela->offset,
+		log_debug("offset %d, type %d, %s %s %ld", rela->offset,
 			rela->type, rela->sym->name,
-			(rela->addend < 0)?"-":"+", abs(rela->addend));
+			(rela->addend < 0)?"-":"+", labs(rela->addend));
 		if (rela->string)
 			log_debug(" (string = %s)", rela->string);
 		log_debug("\n");
 	}
 }
 
-void kpatch_create_section_list(struct kpatch_elf *kelf)
+static void kpatch_create_section_list(struct kpatch_elf *kelf)
 {
 	Elf_Scn *scn = NULL;
 	struct section *sec;
@@ -247,7 +381,11 @@ void kpatch_create_section_list(struct kpatch_elf *kelf)
 		if (!sec->data)
 			ERROR("elf_getdata");
 
-		sec->index = elf_ndxscn(scn);
+		sec->index = (unsigned int)elf_ndxscn(scn);
+
+
+		if (sec->sh.sh_type == SHT_SYMTAB_SHNDX)
+			kelf->symtab_shndx = sec->data;
 
 		log_debug("ndx %02d, data %p, size %zu, name %s\n",
 			sec->index, sec->data->d_buf, sec->data->d_size,
@@ -259,23 +397,117 @@ void kpatch_create_section_list(struct kpatch_elf *kelf)
 		ERROR("expected NULL");
 }
 
-void kpatch_create_symbol_list(struct kpatch_elf *kelf)
+/*
+ * Some x86 kernels have NOP function padding [1] for which objtool [2]
+ * adds ELF function symbols with prefix "__pfx_" to indicate the start
+ * of a function, inclusive of NOP-padding.  Find the prefix symbols and
+ * link them to their corresponding function symbols at an expected
+ * offset.
+ *
+ * A few examples:
+ *
+ *    Value          Size Type    Bind   Vis      Ndx Name
+ * (fork.o, simple case)
+ * 0000000000000000     0 SECTION LOCAL  DEFAULT   31 .text.get_task_mm
+ * 0000000000000000    16 FUNC    GLOBAL DEFAULT   31 __pfx_get_task_mm
+ * 0000000000000010    91 FUNC    GLOBAL DEFAULT   31 get_task_mm
+ *
+ * (fork.o, multiple function aliases)
+ * 0000000000000000     0 SECTION LOCAL  DEFAULT  190 .text.__do_sys_fork
+ * 0000000000000000    16 FUNC    GLOBAL DEFAULT  190 __pfx___x64_sys_fork
+ * 0000000000000010    49 FUNC    LOCAL  DEFAULT  190 __do_sys_fork
+ * 0000000000000010    49 FUNC    GLOBAL DEFAULT  190 __ia32_sys_fork
+ * 0000000000000010    49 FUNC    GLOBAL DEFAULT  190 __x64_sys_fork
+ *
+ * (fork.o multiple functions in one section)
+ * 0000000000000000     0 SECTION LOCAL  DEFAULT   59 .init.text
+ * 0000000000000000    16 FUNC    LOCAL  DEFAULT   59 __pfx_coredump_filter_setup
+ * 0000000000000010    40 FUNC    LOCAL  DEFAULT   59 coredump_filter_setup
+ * 0000000000000038    16 FUNC    WEAK   DEFAULT   59 __pfx_arch_task_cache_init
+ * 0000000000000048    10 FUNC    WEAK   DEFAULT   59 arch_task_cache_init
+ * 0000000000000052    16 FUNC    GLOBAL DEFAULT   59 __pfx_fork_init
+ * 0000000000000062   357 FUNC    GLOBAL DEFAULT   59 fork_init
+ * 00000000000001c7    16 FUNC    GLOBAL DEFAULT   59 __pfx_fork_idle
+ * 00000000000001d7   214 FUNC    GLOBAL DEFAULT   59 fork_idle
+ * 00000000000002ad    16 FUNC    GLOBAL DEFAULT   59 __pfx_mm_cache_init
+ * 00000000000002bd    72 FUNC    GLOBAL DEFAULT   59 mm_cache_init
+ * 0000000000000305    16 FUNC    GLOBAL DEFAULT   59 __pfx_proc_caches_init
+ * 0000000000000315   192 FUNC    GLOBAL DEFAULT   59 proc_caches_init
+ *
+ * (fork.o, function without nop padding / __pfx_ symbol)
+ * 0000000000000000     0 SECTION LOCAL  DEFAULT   99 .text.unlikely.__mmdrop
+ * 0000000000000000    48 FUNC    LOCAL  DEFAULT   99 __mmdrop.cold
+ *
+ * (kpatch-build generated tmp.ko, multple functions in one section, no __pfx_ symbols)
+ * 0000000000000000     0 SECTION LOCAL  DEFAULT   10 .text.unlikely.callback_info.isra.0
+ * 0000000000000010    65 FUNC    LOCAL  DEFAULT   10 callback_info.isra.0
+ * 0000000000000061    54 FUNC    LOCAL  DEFAULT   10 callback_info.isra.0
+ * 00000000000000a7    54 FUNC    LOCAL  DEFAULT   10 callback_info.isra.0
+ *
+ * CONFIG_CFI_CLANG uses something very similar, except the symbol is created
+ * by the compiler and its prefix is "__cfi_".
+ *
+ * [1] bea75b33895f ("x86/Kconfig: Introduce function padding")
+ * [2] 9f2899fe36a6 ("objtool: Add option to generate prefix symbols")
+ */
+static void kpatch_link_prefixed_functions(struct kpatch_elf *kelf)
+{
+	struct symbol *func, *pfx;
+	bool found;
+
+	if (kelf->arch != X86_64)
+		return;
+
+	list_for_each_entry(pfx, &kelf->symbols, list) {
+		if (!pfx->name || pfx->type != STT_FUNC)
+			continue;
+
+		if (strncmp(pfx->name, "__pfx_", 6) &&
+		    strncmp(pfx->name, "__cfi_", 6))
+			continue;
+
+		found = false;
+
+		list_for_each_entry(func, &kelf->symbols, list) {
+			if (func->type == STT_FUNC && func->sec == pfx->sec &&
+			    func->sym.st_value == pfx->sym.st_value + 16) {
+
+				/*
+				 * If a func has aliases, it's possible for
+				 * multiple functions to have the same 'pfx'.
+				 */
+
+				pfx->is_pfx = true;
+				func->pfx = pfx;
+				found = true;
+			}
+		}
+
+		if (!found)
+			ERROR("missing func for %s", pfx->name);
+	}
+}
+
+static void kpatch_create_symbol_list(struct kpatch_elf *kelf)
 {
 	struct section *symtab;
 	struct symbol *sym;
-	int symbols_nr, index = 0;
+	unsigned int symbols_nr, index = 0;
+	Elf32_Word shndx;
 
 	symtab = find_section_by_name(&kelf->sections, ".symtab");
 	if (!symtab)
 		ERROR("missing symbol table");
 
-	symbols_nr = symtab->sh.sh_size / symtab->sh.sh_entsize;
+	symbols_nr = (unsigned int)(symtab->sh.sh_size / symtab->sh.sh_entsize);
 
 	log_debug("\n=== symbol list (%d entries) ===\n", symbols_nr);
 
 	while (symbols_nr--) {
 		ALLOC_LINK(sym, &kelf->symbols);
 
+		INIT_LIST_HEAD(&sym->children);
+
 		sym->index = index;
 		if (!gelf_getsym(symtab->data, index, &sym->sym))
 			ERROR("gelf_getsym");
@@ -289,10 +521,16 @@ void kpatch_create_symbol_list(struct kpatch_elf *kelf)
 		sym->type = GELF_ST_TYPE(sym->sym.st_info);
 		sym->bind = GELF_ST_BIND(sym->sym.st_info);
 
-		if (sym->sym.st_shndx > SHN_UNDEF &&
-		    sym->sym.st_shndx < SHN_LORESERVE) {
-			sym->sec = find_section_by_index(&kelf->sections,
-					sym->sym.st_shndx);
+		shndx = sym->sym.st_shndx;
+		if (shndx == SHN_XINDEX &&
+		    !gelf_getsymshndx(symtab->data, kelf->symtab_shndx, sym->index, &sym->sym, &shndx))
+			ERROR("couldn't find extended section index for symbol %s; idx=%d",
+			      sym->name, sym->index);
+
+		if ((sym->sym.st_shndx > SHN_UNDEF &&
+		    sym->sym.st_shndx < SHN_LORESERVE) ||
+		    sym->sym.st_shndx == SHN_XINDEX) {
+			sym->sec = find_section_by_index(&kelf->sections, shndx);
 			if (!sym->sec)
 				ERROR("couldn't find section for symbol %s\n",
 					sym->name);
@@ -312,33 +550,7 @@ void kpatch_create_symbol_list(struct kpatch_elf *kelf)
 		log_debug("\n");
 	}
 
-}
-
-/* Check which functions have fentry/mcount calls; save this info for later use. */
-static void kpatch_find_func_profiling_calls(struct kpatch_elf *kelf)
-{
-	struct symbol *sym;
-	struct rela *rela;
-	list_for_each_entry(sym, &kelf->symbols, list) {
-		if (sym->type != STT_FUNC || !sym->sec || !sym->sec->rela)
-			continue;
-#ifdef __powerpc__
-		list_for_each_entry(rela, &sym->sec->rela->relas, list) {
-			if (!strcmp(rela->sym->name, "_mcount")) {
-				sym->has_func_profiling = 1;
-				break;
-			}
-		}
-#else
-		rela = list_first_entry(&sym->sec->rela->relas, struct rela,
-					list);
-		if (rela->type != R_X86_64_NONE ||
-		    strcmp(rela->sym->name, "__fentry__"))
-			continue;
-
-		sym->has_func_profiling = 1;
-#endif
-	}
+	kpatch_link_prefixed_functions(kelf);
 }
 
 struct kpatch_elf *kpatch_elf_open(const char *name)
@@ -346,7 +558,8 @@ struct kpatch_elf *kpatch_elf_open(const char *name)
 	Elf *elf;
 	int fd;
 	struct kpatch_elf *kelf;
-	struct section *sec;
+	struct section *relasec;
+	GElf_Ehdr ehdr;
 
 	fd = open(name, O_RDONLY);
 	if (fd == -1)
@@ -367,18 +580,34 @@ struct kpatch_elf *kpatch_elf_open(const char *name)
 	/* read and store section, symbol entries from file */
 	kelf->elf = elf;
 	kelf->fd = fd;
+
+	if (!gelf_getehdr(kelf->elf, &ehdr))
+		ERROR("gelf_getehdr");
+	switch (ehdr.e_machine) {
+	case EM_PPC64:
+		kelf->arch = PPC64;
+		break;
+	case EM_X86_64:
+		kelf->arch = X86_64;
+		break;
+	case EM_S390:
+		kelf->arch = S390;
+		break;
+	default:
+		ERROR("Unsupported target architecture");
+	}
+
 	kpatch_create_section_list(kelf);
 	kpatch_create_symbol_list(kelf);
 
 	/* for each rela section, read and store the rela entries */
-	list_for_each_entry(sec, &kelf->sections, list) {
-		if (!is_rela_section(sec))
+	list_for_each_entry(relasec, &kelf->sections, list) {
+		if (!is_rela_section(relasec))
 			continue;
-		INIT_LIST_HEAD(&sec->relas);
-		kpatch_create_rela_list(kelf, sec);
+		INIT_LIST_HEAD(&relasec->relas);
+		kpatch_create_rela_list(kelf, relasec);
 	}
 
-	kpatch_find_func_profiling_calls(kelf);
 	return kelf;
 }
 
@@ -401,11 +630,11 @@ void kpatch_dump_kelf(struct kpatch_elf *kelf)
 				goto next;
 			printf("rela section expansion\n");
 			list_for_each_entry(rela, &sec->relas, list) {
-				printf("sym %d, offset %d, type %d, %s %s %d\n",
+				printf("sym %d, offset %d, type %d, %s %s %ld\n",
 				       rela->sym->index, rela->offset,
 				       rela->type, rela->sym->name,
 				       (rela->addend < 0)?"-":"+",
-				       abs(rela->addend));
+				       labs(rela->addend));
 			}
 		} else {
 			if (sec->sym)
@@ -430,29 +659,41 @@ next:
 	}
 }
 
-int is_null_sym(struct symbol *sym)
+bool is_null_sym(struct symbol *sym)
 {
 	return !strlen(sym->name);
 }
 
-int is_file_sym(struct symbol *sym)
+bool is_file_sym(struct symbol *sym)
 {
 	return sym->type == STT_FILE;
 }
 
-int is_local_func_sym(struct symbol *sym)
+bool is_local_func_sym(struct symbol *sym)
 {
 	return sym->bind == STB_LOCAL && sym->type == STT_FUNC;
 }
 
-int is_local_sym(struct symbol *sym)
+bool is_local_sym(struct symbol *sym)
 {
 	return sym->bind == STB_LOCAL;
 }
 
+bool is_ubsan_sec(const char *name) {
+	if (!strncmp(name, ".data.rel.local..Lubsan_data", 28) ||
+		!strncmp(name, ".data..Lubsan_type", 18) ||
+		!strncmp(name, ".Lubsan_data", 12) ||
+		!strncmp(name, ".data..Lubsan_data", 18) ||
+		!strncmp(name, ".rela.data..Lubsan_data", 23) ||
+		!strncmp(name, ".rela.data.rel.local..Lubsan_data", 33))
+		return true;
+	else
+		return false;
+}
+
 void print_strtab(char *buf, size_t size)
 {
-	int i;
+	size_t i;
 
 	for (i = 0; i < size; i++) {
 		if (buf[i] == 0)
@@ -470,7 +711,7 @@ void kpatch_create_shstrtab(struct kpatch_elf *kelf)
 
 	shstrtab = find_section_by_name(&kelf->sections, ".shstrtab");
 	if (!shstrtab)
-		ERROR("find_section_by_name");
+		return;
 
 	/* determine size of string table */
 	size = 1; /* for initial NULL terminator */
@@ -487,7 +728,7 @@ void kpatch_create_shstrtab(struct kpatch_elf *kelf)
 	offset = 1;
 	list_for_each_entry(sec, &kelf->sections, list) {
 		len = strlen(sec->name) + 1;
-		sec->sh.sh_name = offset;
+		sec->sh.sh_name = (unsigned int)offset;
 		memcpy(buf + offset, sec->name, len);
 		offset += len;
 	}
@@ -511,7 +752,7 @@ void kpatch_create_shstrtab(struct kpatch_elf *kelf)
 
 void kpatch_create_strtab(struct kpatch_elf *kelf)
 {
-	struct section *strtab;
+	struct section *strtab, *shstrtab;
 	struct symbol *sym;
 	size_t size = 0, offset = 0, len;
 	char *buf;
@@ -520,6 +761,8 @@ void kpatch_create_strtab(struct kpatch_elf *kelf)
 	if (!strtab)
 		ERROR("find_section_by_name");
 
+	shstrtab = find_section_by_name(&kelf->sections, ".shstrtab");
+
 	/* determine size of string table */
 	list_for_each_entry(sym, &kelf->symbols, list) {
 		if (sym->type == STT_SECTION)
@@ -527,6 +770,15 @@ void kpatch_create_strtab(struct kpatch_elf *kelf)
 		size += strlen(sym->name) + 1; /* include NULL terminator */
 	}
 
+	/* and when covering for missing .shstrtab ... */
+	if (!shstrtab) {
+		/* factor out into common (sh)strtab feeder */
+		struct section *sec;
+
+		list_for_each_entry(sec, &kelf->sections, list)
+			size += strlen(sec->name) + 1; /* include NULL terminator */
+	}
+
 	/* allocate data buffer */
 	buf = malloc(size);
 	if (!buf)
@@ -540,13 +792,25 @@ void kpatch_create_strtab(struct kpatch_elf *kelf)
 			continue;
 		}
 		len = strlen(sym->name) + 1;
-		sym->sym.st_name = offset;
+		sym->sym.st_name = (unsigned int)offset;
 		memcpy(buf + offset, sym->name, len);
 		offset += len;
 	}
 
+	if (!shstrtab) {
+		struct section *sec;
+
+		/* populate string table and link with section header */
+		list_for_each_entry(sec, &kelf->sections, list) {
+			len = strlen(sec->name) + 1;
+			sec->sh.sh_name = (unsigned int)offset;
+			memcpy(buf + offset, sec->name, len);
+			offset += len;
+		}
+	}
+
 	if (offset != size)
-		ERROR("shstrtab size mismatch");
+		ERROR("strtab size mismatch");
 
 	strtab->data->d_buf = buf;
 	strtab->data->d_size = size;
@@ -565,10 +829,12 @@ void kpatch_create_strtab(struct kpatch_elf *kelf)
 void kpatch_create_symtab(struct kpatch_elf *kelf)
 {
 	struct section *symtab;
+	struct section *strtab;
 	struct symbol *sym;
 	char *buf;
 	size_t size;
-	int nr = 0, offset = 0, nr_local = 0;
+	int nr = 0, nr_local = 0;
+	unsigned long offset = 0;
 
 	symtab = find_section_by_name(&kelf->sections, ".symtab");
 	if (!symtab)
@@ -598,7 +864,11 @@ void kpatch_create_symtab(struct kpatch_elf *kelf)
 	symtab->data->d_size = size;
 
 	/* update symtab section header */
-	symtab->sh.sh_link = find_section_by_name(&kelf->sections, ".strtab")->index;
+	strtab = find_section_by_name(&kelf->sections, ".strtab");
+	if (!strtab)
+		ERROR("missing .strtab section");
+
+	symtab->sh.sh_link = strtab->index;
 	symtab->sh.sh_info = nr_local;
 }
 
@@ -647,6 +917,7 @@ struct section *create_section_pair(struct kpatch_elf *kelf, char *name,
 	relasec->data = malloc(sizeof(*relasec->data));
 	if (!relasec->data)
 		ERROR("malloc");
+	relasec->data->d_type = ELF_T_RELA;
 
 	/* set section header */
 	relasec->sh.sh_type = SHT_RELA;
@@ -697,7 +968,7 @@ void kpatch_reindex_elements(struct kpatch_elf *kelf)
 {
 	struct section *sec;
 	struct symbol *sym;
-	int index;
+	unsigned int index;
 
 	index = 1; /* elf write function handles NULL section 0 */
 	list_for_each_entry(sec, &kelf->sections, list)
@@ -707,7 +978,7 @@ void kpatch_reindex_elements(struct kpatch_elf *kelf)
 	list_for_each_entry(sym, &kelf->symbols, list) {
 		sym->index = index++;
 		if (sym->sec)
-			sym->sym.st_shndx = sym->sec->index;
+			sym->sym.st_shndx = (unsigned short)sym->sec->index;
 		else if (sym->sym.st_shndx != SHN_ABS &&
 			 sym->sym.st_shndx != SHN_LIVEPATCH)
 			sym->sym.st_shndx = SHN_UNDEF;
@@ -717,8 +988,9 @@ void kpatch_reindex_elements(struct kpatch_elf *kelf)
 void kpatch_rebuild_rela_section_data(struct section *sec)
 {
 	struct rela *rela;
-	int nr = 0, index = 0, size;
+	int nr = 0, index = 0;
 	GElf_Rela *relas;
+	size_t size;
 
 	list_for_each_entry(rela, &sec->relas, list)
 		nr++;
@@ -746,18 +1018,19 @@ void kpatch_rebuild_rela_section_data(struct section *sec)
 		ERROR("size mismatch in rebuilt rela section");
 }
 
-void kpatch_write_output_elf(struct kpatch_elf *kelf, Elf *elf, char *outfile)
+void kpatch_write_output_elf(struct kpatch_elf *kelf, Elf *elf, char *outfile,
+			     mode_t mode)
 {
 	int fd;
 	struct section *sec;
+	struct section *shstrtab;
 	Elf *elfout;
 	GElf_Ehdr eh, ehout;
 	Elf_Scn *scn;
 	Elf_Data *data;
 	GElf_Shdr sh;
 
-	/* TODO make this argv */
-	fd = creat(outfile, 0777);
+	fd = creat(outfile, mode);
 	if (fd == -1)
 		ERROR("creat");
 
@@ -765,7 +1038,7 @@ void kpatch_write_output_elf(struct kpatch_elf *kelf, Elf *elf, char *outfile)
 	if (!elfout)
 		ERROR("elf_begin");
 
-	if (!gelf_newehdr(elfout, gelf_getclass(kelf->elf)))
+	if (!gelf_newehdr(elfout, gelf_getclass(elf)))
 		ERROR("gelf_newehdr");
 
 	if (!gelf_getehdr(elfout, &ehout))
@@ -779,7 +1052,15 @@ void kpatch_write_output_elf(struct kpatch_elf *kelf, Elf *elf, char *outfile)
 	ehout.e_machine = eh.e_machine;
 	ehout.e_type = eh.e_type;
 	ehout.e_version = EV_CURRENT;
-	ehout.e_shstrndx = find_section_by_name(&kelf->sections, ".shstrtab")->index;
+	ehout.e_flags = eh.e_flags;
+
+	shstrtab = find_section_by_name(&kelf->sections, ".shstrtab");
+	if (!shstrtab)
+		shstrtab = find_section_by_name(&kelf->sections, ".strtab");
+	if (!shstrtab)
+		ERROR("missing .shstrtab, .strtab sections");
+
+	ehout.e_shstrndx = (unsigned short)shstrtab->index;
 
 	/* add changed sections */
 	list_for_each_entry(sec, &kelf->sections, list) {
@@ -814,6 +1095,9 @@ void kpatch_write_output_elf(struct kpatch_elf *kelf, Elf *elf, char *outfile)
 		printf("%s\n",elf_errmsg(-1));
 		ERROR("elf_update");
 	}
+
+	elf_end(elfout);
+	close(fd);
 }
 
 /*
@@ -830,17 +1114,21 @@ void kpatch_elf_teardown(struct kpatch_elf *kelf)
 	struct rela *rela, *saferela;
 
 	list_for_each_entry_safe(sec, safesec, &kelf->sections, list) {
+		if (sec->twin)
+			sec->twin->twin = NULL;
 		if (is_rela_section(sec)) {
 			list_for_each_entry_safe(rela, saferela, &sec->relas, list) {
 				memset(rela, 0, sizeof(*rela));
 				free(rela);
 			}
-			memset(sec, 0, sizeof(*sec));
-			free(sec);
 		}
+		memset(sec, 0, sizeof(*sec));
+		free(sec);
 	}
 
 	list_for_each_entry_safe(sym, safesym, &kelf->symbols, list) {
+		if (sym->twin)
+			sym->twin->twin = NULL;
 		memset(sym, 0, sizeof(*sym));
 		free(sym);
 	}
