@@ -77,6 +77,7 @@ struct symbol {
 	struct list_head list;
 	struct symbol *twin;
 	struct symbol *parent;
+	struct symbol *pfx;
 	struct list_head children;
 	struct list_head subfunction_node;
 	struct section *sec;
@@ -91,6 +92,7 @@ struct symbol {
 		enum symbol_strip strip; /* used in the output elf */
 	};
 	int has_func_profiling;
+	bool is_pfx;
 };
 
 struct rela {
@@ -170,6 +172,7 @@ bool is_null_sym(struct symbol *sym);
 bool is_file_sym(struct symbol *sym);
 bool is_local_func_sym(struct symbol *sym);
 bool is_local_sym(struct symbol *sym);
+bool is_ubsan_sec(const char *name);
 
 void print_strtab(char *buf, size_t size);
 void kpatch_create_shstrtab(struct kpatch_elf *kelf);
